﻿// The sole owner of this code is Dundas Data Visualization Inc.
// Copyright (c) 2015 Dundas Data Visualization Inc.
// All rights reserved.

using Dundas.BI;
using Dundas.BI.AccountServices.Extensibility;
using Dundas.BI.AccountServices.MultiTenancy;
using Dundas.BI.Licensing;
using Dundas.BI.Utility;

using Newtonsoft.Json;

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;

namespace MyCompany.Sample.CustomTenantsProviderSample
{
    /// <summary>
    /// This class represents a simple tenant provider that uses a csv file.
    /// </summary>
	public class CustomTenantsProvider : ITenantsProvider
	{
		#region Fields

		private string _tenantInformationFilePath;

		#endregion Fields

		#region Internal Properties

		/// <summary>
		/// Gets the tenant information file path.
		/// </summary>
		/// <value>
		/// The tenant information file path.
		/// </value>
		internal string TenantInformationFilePath
		{
			get
			{
				if(_tenantInformationFilePath == null)
				{
					_tenantInformationFilePath =
						System.Web.HttpContext.Current.Server.MapPath("/App_Data/TenantInformation.csv");
				}

				return _tenantInformationFilePath;
			}
		}

		#endregion Internal Properties

		#region Public Properties

        /// <summary>
        /// Gets a value indicating whether the provider supports saving changes.
        /// </summary>
        public bool IsModifySupported
        {
	        get { return true; }
        }

		/// <summary>
		/// Gets the query operations which are supported by the provider.
		/// </summary>
		public QueryApiOperations SupportedQueryOperations
		{
			get { return QueryApiOperations.Count; }
		}

		#endregion Public Properties

		#region Private Methods

		/// <summary>
		/// Deserializes the license seat allocation.
		/// </summary>
		/// <param name="licenseSeatAllocationDataAsString">The license seat allocation data as string.</param>
		/// <returns>A TenantSeatAllocationData object.</returns>
		private TenantSeatAllocationData DeserializeLicenseSeatAllocation(string licenseSeatAllocationDataAsString)
		{
			return JsonConvert.DeserializeObject<TenantSeatAllocationData>(licenseSeatAllocationDataAsString, new JsonSerializerSettings());
		}

        /// <summary>
        /// Saves all tenant data to file.
        /// </summary>
        /// <param name="tenantDataList">The tenant data list.</param>
        private void SaveAllTenantDataToFile(IEnumerable<TenantData> tenantDataList)
        {
	        StringBuilder updatedFileContent = new StringBuilder();

	        foreach(TenantData tenantData in tenantDataList)
	        {
		        updatedFileContent.AppendFormat(
			        CultureInfo.CurrentCulture,
			        "{0}•{1}•{2}•{3}•{4}•{5}•{6}•{7}•{8}{9}",
			        tenantData.Id,
			        tenantData.Name,
			        tenantData.AdministratorsGroupId,
			        tenantData.CreatedTime,
			        tenantData.MembersGroupId,
			        tenantData.AccountNamePattern,
			        tenantData.BrandingSettingOverrides,
			        tenantData.DataConnectorPropertyOverrides,
			        SerializeLicenseSeatAllocation(tenantData.LicenseSeatAllocation),
			        Environment.NewLine
			        );
	        }

	        File.WriteAllText(this.TenantInformationFilePath, updatedFileContent.ToString());
        }

		/// <summary>
		/// Serializes the license seat allocation.
		/// </summary>
		/// <param name="tenantSeatAllocationData">The tenant seat allocation data.</param>
		/// <returns>A string representing a tenantSeatAllocation Data object.</returns>
		private string SerializeLicenseSeatAllocation(TenantSeatAllocationData tenantSeatAllocationData)
		{
			return JsonConvert.SerializeObject(tenantSeatAllocationData);
		}

		#endregion Private Methods

		#region Public Methods

        /// <summary>
        /// Deletes a tenant specified by its ID.
        /// </summary>
        /// <param name="tenantId">The tenant ID.</param>
        public void DeleteRecord(Guid tenantId)
        {
	        SaveAllTenantDataToFile(
		        GetAllRecords().Where(
			        record => !record.Id.Equals(tenantId)
			        )
	        );
        }

        /// <summary>
        /// Gets all records.
        /// </summary>
        /// <returns>
        /// All the records managed by the provider.
        /// </returns>
        public IEnumerable<TenantData> GetAllRecords()
        {
	        List<TenantData> tenantDataList =
		        new List<TenantData>();

	        foreach(string line in File.ReadAllLines(
		        this.TenantInformationFilePath
		        )
	        )
	        {
		        string[] splitLine =
			        line.Split('•');

		        tenantDataList.Add(
			        new TenantData()
			        {
					        Id = new Guid(splitLine[0]),
					        Name = splitLine[1],
					        AdministratorsGroupId = new Guid(splitLine[2]),
					        CreatedTime = DateTime.Parse(splitLine[3]),
					        MembersGroupId = new Guid(splitLine[4]),
					        AccountNamePattern = splitLine[5],
					        BrandingSettingOverrides = splitLine[6],
					        DataConnectorPropertyOverrides = splitLine[7],
					        LicenseSeatAllocation = DeserializeLicenseSeatAllocation(splitLine[8])
			        }
		        );
	        }

	        return tenantDataList;
        }

        /// <summary>
        /// Retrieves a list of tenants having the specified IDs.
        /// </summary>
        /// <param name="tenantIds">The IDs of the tenants to retrieve.</param>
        /// <returns>
        /// The requested tenants.
        /// </returns>
        public ICollection<TenantData> GetById(ICollection<Guid> tenantIds)
        {
	        Collection<TenantData> tenantDataCollection =
		        new Collection<TenantData>();

	        var allRecords =
		        GetAllRecords();

	        foreach (Guid id in tenantIds)
	        {
		        TenantData foundItem =
			        allRecords.FirstOrDefault(
				        accountDataItem => accountDataItem.Id.Equals(id)
			        );

		        if (foundItem != null)
		        {
			        tenantDataCollection.Add(
				        foundItem
				        );
		        }
	        }

	        return tenantDataCollection;
        }

        /// <summary>
        /// Gets the IDs of any tenants which draw the specified seat kind from the global pool.
        /// </summary>
        /// <param name="seatKind">The kind of license seat.</param>
        /// <param name="seatCountProperty">The seat count property.</param>
        /// <returns>
        /// IDs of the tenant which draw from global pool.
        /// </returns>
        public ICollection<Guid> GetIdsOfTenantsWhichDrawFromGlobalPool(LicenseSeatKind2 seatKind, SeatCountProperty seatCountProperty)
        {
	        // Only Tenant1 draws from global pool in this example.
	        return new Collection<Guid>()
		        {
			        GetAllRecords().FirstOrDefault(tenantItem => tenantItem.Name.Equals("tenant1")).Id
		        };
        }

        /// <summary>
        /// Gets the combined license usage for all tenants.
        /// </summary>
        /// <returns>
        /// A dictionary containing the total number of seats allocated to all tenants, grouped by seat kind.
        /// </returns>
        public IDictionary<LicenseSeatKind2, SeatCount> GetLicenseAllocation()
        {
	        return new Dictionary<LicenseSeatKind2, SeatCount>()
	        {
		        { LicenseSeatKind2.ReservedDeveloper, new SeatCount(10) },
		        { LicenseSeatKind2.ReservedPowerUser, new SeatCount(10) },
		        { LicenseSeatKind2.ReservedStandardUser, new SeatCount(10) }
	        };
        }

        /// <summary>
        /// Imports a tenant to the underlying storage mechanism.
        /// </summary>
        /// <param name="tenant">The tenant to be imported.</param>
        /// <returns>
        ///   <see langword="true" /> if a new tenant was imported; otherwise <see langword="false" />.
        /// </returns>
        public bool ImportRecord(TenantData tenant)
        {
	        IEnumerable<TenantData> allTenants =
		        GetAllRecords();  

	        TenantData tenantData =
		        allTenants.FirstOrDefault(
			        tenantItem =>
				        tenantItem.Id.Equals(tenant.Id)
		        );

	        if(tenantData != null)
	        {
		        // Overwrite existin record.
		        tenantData = tenant;

		        SaveAllTenantDataToFile(allTenants);

		        return false;
	        }
	        else
	        {
		        // New record.
		        SaveAllTenantDataToFile(
			        allTenants.Concat(new TenantData[] {  tenant } )
			        );
		        return true;
	        }
        }

		/// <summary>
		/// Queries for tenant records.
		/// </summary>
		/// <param name="pageNumber">The page number.</param>
		/// <param name="pageSize">The number of results in each page.</param>
		/// <param name="orderBy">The sort order of the result, or <see langword="null" /> if the order does not matter.</param>
		/// <param name="filter">The filter rules which should be applied to the query, or <see langword="null" /> if no filters are required.</param>
		/// <returns>
		/// The records matching the search criteria.
		/// </returns>
		/// <exception cref="System.NotSupportedException"></exception>
		public IList<Guid> Query(int pageNumber, int pageSize, IList<Tuple<TenantQueryField, SortDirection>> orderBy, ICollection<TenantQueryFilterRule> filter)
		{
			throw new NotSupportedException();
		}

		/// <summary>
		/// Queries for the number of tenants matching a filter criteria.
		/// </summary>
		/// <param name="filter">The filter rules which should be applied to the query, or <see langword="null" /> if
		/// the total number of tenants should be returned.</param>
		/// <returns>
		/// The number of tenants matching the filter.
		/// </returns>
		public int QueryCount(ICollection<TenantQueryFilterRule> filter)
		{
			return GetAllRecords().Count();
		}

        /// <summary>
        /// Saves a tenant.
        /// </summary>
        /// <param name="tenant">The tenant to be saved.</param>
        /// <returns>
        /// The ID of the tenant.
        /// </returns>
        public Guid SaveRecord(TenantData tenant)
        {
	        TenantData[] listToSave;

	        if(tenant.Id.Equals(Guid.Empty))
	        {
		        // Record is new
		        tenant.Id = Guid.NewGuid();
		        tenant.CreatedTime = DateTime.Now;

		        listToSave =
			        GetAllRecords().Concat(
				        new TenantData[] { tenant }
			        ).ToArray();
	        }
	        else
	        {
		        // Existing record
		        listToSave =
			        GetAllRecords().ToArray();

		        for (int index = 0; index < listToSave.Count(); index++)
		        {
			        if (listToSave[index].Id.Equals(tenant.Id))
			        {
				        listToSave[index] =
					        tenant;
				        break;
			        }
		        }
	        }

	        SaveAllTenantDataToFile(listToSave);

	        return tenant.Id;
        }

        /// <summary>
        /// Update the special groups associated with the tenant.
        /// </summary>
        /// <remarks>
        /// After a tenant is created, two groups are also created (Tenant Admin, and Member groups).
        /// </remarks>
        /// <param name="tenantId">ID of the tenant.</param>
        /// <param name="adminGroupId">The tenant admin group ID.</param>
        /// <param name="memberGroupId">The tenant group ID.</param>
        public void UpdateSpecialGroupIds(Guid tenantId, Guid adminGroupId, Guid memberGroupId)
        {
	        TenantData[] listToSave =
		        GetAllRecords().ToArray();

	        for (int index = 0; index < listToSave.Count(); index++)
	        {
		        if (listToSave[index].Id.Equals(tenantId))
		        {
			        listToSave[index].AdministratorsGroupId = adminGroupId;
			        listToSave[index].MembersGroupId = memberGroupId;
			        break;
		        }
	        }

	        SaveAllTenantDataToFile(listToSave);
        }

		#endregion Public Methods
	}
}